import random
from typing import Any

from jafgen.stores.item import Item


class Inventory(object):
    inventory: dict[Any, Any] = {}

    @classmethod
    def update(cls, inventory_list):
        cls.inventory["jaffle"] = []
        cls.inventory["beverage"] = []
        for item in inventory_list:
            ttype = item.type
            cls.inventory[ttype].append(item)

    @classmethod
    def get_food(cls, count=1):
        return [random.choice(cls.inventory["jaffle"]) for i in range(count)]

    @classmethod
    def get_drink(cls, count=1):
        return [random.choice(cls.inventory["beverage"]) for i in range(count)]

    @classmethod
    def to_dict(cls) -> list[dict[str, Any]]:
        all_items = []
        for key in cls.inventory:
            all_items += [item.to_dict() for item in cls.inventory[key]]
        return all_items


Inventory.update(
    [
        Item(
            sku="JAF-001",
            name="nutellaphone who dis?",
            description="nutella and banana jaffle",
            type="jaffle",
            price=11,
        ),
        Item(
            sku="JAF-002",
            name="doctor stew",
            description="house-made beef stew jaffle",
            type="jaffle",
            price=11,
        ),
        Item(
            sku="JAF-003",
            name="the krautback",
            description="lamb and pork bratwurst with house-pickled cabbage sauerkraut and mustard",
            type="jaffle",
            price=12,
        ),
        Item(
            sku="JAF-004",
            name="flame impala",
            description="pulled pork and pineapple al pastor marinated in ghost pepper sauce, kevin parker's favorite! ",
            type="jaffle",
            price=14,
        ),
        Item(
            sku="JAF-005",
            name="mel-bun",
            description="melon and minced beef bao, in a jaffle, savory and sweet",
            type="jaffle",
            price=12,
        ),
        Item(
            sku="BEV-001",
            name="tangaroo",
            description="mango and tangerine smoothie",
            type="beverage",
            price=6,
        ),
        Item(
            sku="BEV-002",
            name="chai and mighty",
            description="oatmilk chai latte with protein boost",
            type="beverage",
            price=5,
        ),
        Item(
            sku="BEV-003",
            name="vanilla ice",
            description="iced coffee with house-made french vanilla syrup",
            type="beverage",
            price=6,
        ),
        Item(
            sku="BEV-004",
            name="for richer or pourover ",
            description="daily selection of single estate beans for a delicious hot pourover",
            type="beverage",
            price=7,
        ),
        Item(
            sku="BEV-005",
            name="adele-ade",
            description="a kiwi and lime agua fresca, hello from the other side of thirst",
            type="beverage",
            price=4,
        ),
    ]
)



import random

import numpy as np

from jafgen.customers.customers import (
    BrunchCrowd,
    Casuals,
    Commuter,
    RemoteWorker,
    Student,
    HealthNut,
)


class Market(object):
    PersonaMix = [
        (Commuter, 0.25),
        (RemoteWorker, 0.25),
        (BrunchCrowd, 0.1),
        (Student, 0.2),
        (Casuals, 0.1),
        (HealthNut, 0.1),
    ]

    def __init__(self, store, num_customers, days_to_penetration=365):
        self.store = store
        self.num_customers = num_customers
        self.days_to_penetration = days_to_penetration

        self.addressable_customers = []

        for Persona, weight in self.PersonaMix:
            num_customers = int(weight * self.num_customers)
            for i in range(num_customers):
                self.addressable_customers.append(Persona(store))

        random.shuffle(self.addressable_customers)

        self.active_customers = []

    def sim_day(self, day):
        days_since_open = self.store.days_since_open(day)
        if days_since_open < 0:
            yield None
            return
        elif days_since_open < 7:
            pct_penetration = min(days_since_open / self.days_to_penetration, 1)
            market_penetration = min(np.log(1.2 + pct_penetration * (np.e - 1.2)), 1)
        else:
            pct_penetration = min(days_since_open / self.days_to_penetration, 1)
            market_penetration = min(np.log(1 + pct_penetration * (np.e - 1)), 1)

        num_desired_customers = market_penetration * len(self.addressable_customers)
        customers_to_add = int(num_desired_customers - len(self.active_customers))

        for i in range(customers_to_add):
            customer = self.addressable_customers.pop()
            self.active_customers.append(customer)

        for customer in self.active_customers:
            order, tweet = customer.sim_day(day)
            yield order, tweet



from typing import Any


class Supply(object):
    def __init__(self, id, name, cost, perishable, skus):
        self.id = id
        self.name = name
        self.cost = cost
        self.perishable = perishable
        self.skus = skus

    def __str__(self):
        return f"<{self.name} @ ${self.cost}>"

    def __repr__(self):
        return self.__str__()

    def to_dict(self, sku) -> dict[str, Any]:
        return {
            "id": str(self.id),
            "name": str(self.name),
            "cost": int(self.cost * 100),
            "perishable": str(self.perishable),
            "sku": str(sku),
        }



import uuid
from typing import Any


class OrderItem(object):
    def __init__(self, order_id, item):
        self.item_id = str(uuid.uuid4())
        self.order_id = order_id
        self.item = item

    def to_dict(self) -> dict[str, Any]:
        return {
            "id": self.item_id,
            "order_id": self.order_id,
            "sku": self.item.sku,
        }



import datetime
from jafgen.stores.store import Store
from jafgen.customers.customers import (
    RemoteWorker,
    BrunchCrowd,
    HealthNut,
    Commuter,
    Casuals,
    Student,
)
from jafgen.curves import Day
from jafgen.simulation import HoursOfOperation

T_7AM = 60 * 7
T_8AM = 60 * 8
T_3PM = 60 * 15
T_8PM = 60 * 20


def test_tweets():
    """Test that tweets only come after orders and in the range of 20 minutes after."""

    store = Store(
        str(1),
        "Testylvania",
        0.85,
        HoursOfOperation(
            weekday_range=(T_7AM, T_8PM),
            weekend_range=(T_8AM, T_3PM),
        ),
        0,
        0.0659123,
    )
    customers = []
    personas = [RemoteWorker, BrunchCrowd, HealthNut, Commuter, Casuals, Student]
    for i in range(100):
        customers.append(personas[i % len(personas)](store))

    for i in range(100):
        day = Day(i)
        for customer in customers:
            order, tweet = customer.sim_day(day)
            if tweet:
                assert tweet.customer == customer
                assert tweet.order == order
                assert (
                    tweet.tweeted_at
                    <= tweet.order.order_time.date + datetime.timedelta(minutes=20)
                )
            if not order:
                assert not tweet



import csv
import os
import uuid

from rich.progress import track

from jafgen.curves import Day
from jafgen.stores.inventory import Inventory
from jafgen.stores.market import Market
from jafgen.stores.stock import Stock
from jafgen.stores.store import Store

T_7AM = 60 * 7
T_8AM = 60 * 8
T_3PM = 60 * 15
T_8PM = 60 * 20


class HoursOfOperation(object):
    def __init__(self, weekday_range, weekend_range):
        self.weekday_range = weekday_range
        self.weekend_range = weekend_range

    def minutes_open(self, date):
        if date.is_weekend:
            return self.weekend_range[1] - self.weekend_range[0]
        else:
            return self.weekday_range[1] - self.weekday_range[0]

    def opens_at(self, date):
        if date.is_weekend:
            return self.weekend_range[0]
        else:
            return self.weekday_range[0]

    def closes_at(self, date):
        if date.is_weekend:
            return self.weekend_range[1]
        else:
            return self.weekday_range[1]

    def is_open(self, date):
        opens_at = self.opens_at(date)
        closes_at = self.closes_at(date)

        dt = date.date.hour * 60 + date.date.minute
        return dt >= opens_at and dt < closes_at

    def iter_minutes(self, date):
        if date.is_weekend:
            start, end = self.weekend_range
        else:
            start, end = self.weekday_range

        for i in range(end - start):
            yield start + i


class Simulation(object):
    def __init__(self, years, prefix):
        self.years = years
        self.scale = 100
        self.prefix = prefix
        self.stores = [
            # id | name | popularity | opened | TAM | tax
            (str(uuid.uuid4()), "Philadelphia", 0.85, 0, 9 * self.scale, 0.06),
            (str(uuid.uuid4()), "Brooklyn", 0.95, 192, 14 * self.scale, 0.04),
            (str(uuid.uuid4()), "Chicago", 0.92, 605, 12 * self.scale, 0.0625),
            (str(uuid.uuid4()), "San Francisco", 0.87, 615, 11 * self.scale, 0.075),
            (str(uuid.uuid4()), "New Orleans", 0.92, 920, 8 * self.scale, 0.04),
            (str(uuid.uuid4()), "Los Angeles", 0.87, 1107, 8 * self.scale, 0.08),
        ]

        self.markets = []
        for (
            store_id,
            store_name,
            popularity,
            opened_date,
            market_size,
            tax,
        ) in self.stores:
            market = Market(
                Store(
                    store_id=store_id,
                    name=store_name,
                    base_popularity=popularity,
                    hours_of_operation=HoursOfOperation(
                        weekday_range=(T_7AM, T_8PM),
                        weekend_range=(T_8AM, T_3PM),
                    ),
                    opened_date=Day(opened_date),
                    tax_rate=tax,
                ),
                num_customers=market_size,
            )
            self.markets.append(market)

        self.customers = {}
        self.orders = []
        self.tweets = []
        self.sim_days = 365 * self.years

    def run_simulation(self):
        for i in track(
            range(self.sim_days), description="🥪 Pressing fresh jaffles..."
        ):
            for market in self.markets:
                day = Day(i)
                for result in market.sim_day(day):
                    if result is not None and all(val is not None for val in result):
                        order, tweet = result
                        self.orders.append(order)
                        self.tweets.append(tweet)
                        if order.customer.customer_id not in self.customers:
                            self.customers[order.customer.customer_id] = order.customer

    def save_results(self) -> None:
        stock: Stock = Stock()
        inventory: Inventory = Inventory()
        entities: dict[str, list[dict]] = {
            "customers": [customer.to_dict() for customer in self.customers.values()],
            "orders": [order.to_dict() for order in self.orders],
            "items": [item.to_dict() for order in self.orders for item in order.items],
            "stores": [market.store.to_dict() for market in self.markets],
            "supplies": stock.to_dict(),
            "products": inventory.to_dict(),
            "tweets": [tweet.to_dict() for tweet in self.tweets],
        }

        if not os.path.exists("./jaffle-data"):
            os.makedirs("./jaffle-data")
        for entity, data in track(
            entities.items(), description="🚚 Delivering jaffles..."
        ):
            with open(
                f"./jaffle-data/{self.prefix}_{entity}.csv", "w", newline=""
            ) as file:
                writer = csv.DictWriter(file, fieldnames=data[0].keys())
                writer.writeheader()
                writer.writerows(data)



from typing import Any


class Store(object):
    def __init__(
        self, store_id, name, base_popularity, hours_of_operation, opened_date, tax_rate
    ):
        self.store_id = store_id
        self.name = name
        self.base_popularity = base_popularity
        self.hours_of_operation = hours_of_operation
        self.opened_date = opened_date
        self.tax_rate = tax_rate

    def p_buy(self, date):
        date_effect = date.get_effect()
        return self.base_popularity * date_effect

    def minutes_open(self, date):
        return self.hours_of_operation.minutes_open(date)

    def iter_minutes_open(self, date):
        yield from self.hours_of_operation.iter_minutes(date)

    def is_open(self, date):
        return date.date >= self.opened_date.date

    def is_open_at(self, date):
        return self.hours_of_operation.is_open(date)

    def days_since_open(self, date):
        return date.date_index - self.opened_date.date_index

    def opens_at(self, date):
        return self.hours_of_operation.opens_at(date)

    def closes_at(self, date):
        return self.hours_of_operation.closes_at(date)

    def to_dict(self) -> dict[str, Any]:
        return {
            "id": str(self.store_id),
            "name": str(self.name),
            "opened_at": str(self.opened_date.date.isoformat()),
            "tax_rate": str(self.tax_rate),
        }



from typing import Any

from jafgen.stores.supply import Supply


class Stock(object):
    stock: dict[Any, Any] = {}

    @classmethod
    def update(cls, stock_list):
        for supply in stock_list:
            skus = supply.skus
            for sku in skus:
                if sku not in cls.stock:
                    cls.stock[sku] = []
                cls.stock[sku].append(supply)

    @classmethod
    def to_dict(cls) -> list[dict[str, Any]]:
        all_items = []
        for key in cls.stock:
            all_items += [item.to_dict(key) for item in cls.stock[key]]
        return all_items


Stock.update(
    [
        Supply(
            id="SUP-001",
            name="compostable cutlery - knife",
            cost=0.07,
            perishable=False,
            skus=["JAF-001", "JAF-002", "JAF-003", "JAF-004", "JAF-005"],
        ),
        Supply(
            id="SUP-002",
            name="cutlery - fork",
            cost=0.07,
            perishable=False,
            skus=["JAF-001", "JAF-002", "JAF-003", "JAF-004", "JAF-005"],
        ),
        Supply(
            id="SUP-003",
            name="serving boat",
            cost=0.11,
            perishable=False,
            skus=["JAF-001", "JAF-002", "JAF-003", "JAF-004", "JAF-005"],
        ),
        Supply(
            id="SUP-004",
            name="napkin",
            cost=0.04,
            perishable=False,
            skus=["JAF-001", "JAF-002", "JAF-003", "JAF-004", "JAF-005"],
        ),
        Supply(
            id="SUP-005",
            name="16oz compostable clear cup",
            cost=0.13,
            perishable=False,
            skus=["BEV-001", "BEV-002", "BEV-003", "BEV-004", "BEV-005"],
        ),
        Supply(
            id="SUP-006",
            name="16oz compostable clear lid",
            cost=0.04,
            perishable=False,
            skus=["BEV-001", "BEV-002", "BEV-003", "BEV-004", "BEV-005"],
        ),
        Supply(
            id="SUP-007",
            name="biodegradable straw",
            cost=0.13,
            perishable=False,
            skus=["BEV-001", "BEV-002", "BEV-003", "BEV-004", "BEV-005"],
        ),
        Supply(
            id="SUP-008", name="chai mix", cost=0.98, perishable=True, skus=["BEV-002"]
        ),
        Supply(
            id="SUP-009",
            name="bread",
            cost=0.33,
            perishable=True,
            skus=["JAF-001", "JAF-002", "JAF-003", "JAF-004", "JAF-005"],
        ),
        Supply(
            id="SUP-010",
            name="cheese",
            cost=0.2,
            perishable=True,
            skus=["JAF-002", "JAF-003", "JAF-004", "JAF-005"],
        ),
        Supply(
            id="SUP-011", name="nutella", cost=0.46, perishable=True, skus=["JAF-001"]
        ),
        Supply(
            id="SUP-012", name="banana", cost=0.13, perishable=True, skus=["JAF-001"]
        ),
        Supply(
            id="SUP-013", name="beef stew", cost=1.69, perishable=True, skus=["JAF-002"]
        ),
        Supply(
            id="SUP-014",
            name="lamb and pork bratwurst",
            cost=2.34,
            perishable=True,
            skus=["JAF-003"],
        ),
        Supply(
            id="SUP-015",
            name="house-pickled cabbage sauerkraut",
            cost=0.43,
            perishable=True,
            skus=["JAF-003"],
        ),
        Supply(
            id="SUP-016", name="mustard", cost=0.07, perishable=True, skus=["JAF-003"]
        ),
        Supply(
            id="SUP-017",
            name="pulled pork",
            cost=2.15,
            perishable=True,
            skus=["JAF-004"],
        ),
        Supply(
            id="SUP-018", name="pineapple", cost=0.26, perishable=True, skus=["JAF-004"]
        ),
        Supply(
            id="SUP-019", name="melon", cost=0.33, perishable=True, skus=["JAF-005"]
        ),
        Supply(
            id="SUP-020",
            name="minced beef",
            cost=1.24,
            perishable=True,
            skus=["JAF-005"],
        ),
        Supply(
            id="SUP-021",
            name="ghost pepper sauce",
            cost=0.2,
            perishable=True,
            skus=["JAF-004"],
        ),
        Supply(
            id="SUP-022", name="mango", cost=0.32, perishable=True, skus=["BEV-001"]
        ),
        Supply(
            id="SUP-023", name="tangerine", cost=0.2, perishable=True, skus=["BEV-001"]
        ),
        Supply(
            id="SUP-024", name="oatmilk", cost=0.11, perishable=True, skus=["BEV-002"]
        ),
        Supply(
            id="SUP-025",
            name="whey protein",
            cost=0.36,
            perishable=True,
            skus=["BEV-002"],
        ),
        Supply(
            id="SUP-026",
            name="coffee",
            cost=0.52,
            perishable=True,
            skus=["BEV-003", "BEV-004"],
        ),
        Supply(
            id="SUP-027",
            name="french vanilla syrup",
            cost=0.72,
            perishable=True,
            skus=["BEV-003"],
        ),
        Supply(id="SUP-028", name="kiwi", cost=0.2, perishable=True, skus=["BEV-005"]),
        Supply(id="SUP-029", name="lime", cost=0.13, perishable=True, skus=["BEV-005"]),
    ]
)



from typing import Any


class Item(object):
    def __init__(self, sku, name, description, type, price):
        self.sku = sku
        self.name = name
        self.description = description
        self.type = type
        self.price = price

    def __str__(self):
        return f"<{self.name} @ ${self.price}>"

    def __repr__(self):
        return self.__str__()

    def to_dict(self) -> dict[str, Any]:
        return {
            "sku": str(self.sku),
            "name": str(self.name),
            "type": str(self.type),
            "price": int(self.price * 100),
            "description": str(self.description),
        }



from jafgen.customers.order import Order
from jafgen.curves import Day
from jafgen.stores.store import Store
from jafgen.customers.customers import RemoteWorker, BrunchCrowd, Student
from jafgen.stores.inventory import Inventory


def test_order_totals():
    """Ensure order totals are equivalent to the sum of the item prices and tax paid"""

    store = Store(str(1), "Testylvania", 0.85, 0, 9 * 100, 0.0659123)
    inventory = Inventory()
    orders = []
    for i in range(1000):
        orders.append(
            Order(
                customer=RemoteWorker(store=store),
                items=[
                    inventory.get_food()[0],
                    inventory.get_drink()[0],
                    inventory.get_food()[0],
                ],
                store=store,
                order_time=Day(date_index=i),
            )
        )
        orders.append(
            Order(
                customer=BrunchCrowd(store=store),
                items=[
                    inventory.get_food()[0],
                    inventory.get_drink()[0],
                    inventory.get_food()[0],
                ],
                store=store,
                order_time=Day(date_index=i),
            )
        )
        orders.append(
            Order(
                customer=Student(store=store),
                items=[
                    inventory.get_food()[0],
                    inventory.get_drink()[0],
                    inventory.get_food()[0],
                ],
                store=store,
                order_time=Day(date_index=i),
            )
        )
    for order in orders:
        assert (
            order.subtotal
            == order.items[0].item.price
            + order.items[1].item.price
            + order.items[2].item.price
        )
        assert order.tax_paid == order.subtotal * order.store.tax_rate
        assert order.order_total == order.subtotal + order.tax_paid
        assert round(float(order.order_total), 2) == round(
            float(order.subtotal), 2
        ) + round(float(order.tax_paid), 2)
        order_dict = order.to_dict()
        assert (
            order_dict["order_total"] == order_dict["subtotal"] + order_dict["tax_paid"]
        )



import datetime
from abc import ABC, abstractmethod
import numpy as np


class Curve(ABC):
    @property
    @abstractmethod
    def Domain(self):
        raise NotImplementedError

    @abstractmethod
    def TranslateDomain(self, date):
        raise NotImplementedError

    @abstractmethod
    def Expr(self, x):
        raise NotImplementedError

    @classmethod
    def eval(cls, date):
        instance = cls()
        domain_value = instance.TranslateDomain(date)
        domain_index = domain_value % len(instance.Domain)
        translated_value = instance.Domain[domain_index]
        return instance.Expr(translated_value)


class AnnualCurve(Curve):
    @property
    def Domain(self):
        return np.linspace(0, 2 * np.pi, 365)

    def TranslateDomain(self, date):
        return date.timetuple().tm_yday

    def Expr(self, x):
        return (np.cos(x) + 1) / 10 + 0.8


class WeekendCurve(Curve):
    @property
    def Domain(self):
        return tuple(range(6))

    def TranslateDomain(self, date):
        return date.weekday() - 1

    def Expr(self, x):
        if x >= 6:
            return 0.6
        else:
            return 1.0


class GrowthCurve(Curve):
    @property
    def Domain(self):
        return tuple(range(500))

    def TranslateDomain(self, date):
        return (date.year - 2016) * 12 + date.month

    def Expr(self, x):
        # ~ aim for ~20% growth/year
        return 1 + (x / 12) * 0.2


class Day(object):
    EPOCH = datetime.datetime(year=2018, month=9, day=1)
    SEASONAL_MONTHLY_CURVE = AnnualCurve()
    WEEKEND_CURVE = WeekendCurve()
    GROWTH_CURVE = GrowthCurve()

    def __init__(self, date_index, minutes=0):
        self.date_index = date_index
        self.date = self.EPOCH + datetime.timedelta(days=date_index, minutes=minutes)
        self.effects = [
            self.SEASONAL_MONTHLY_CURVE.eval(self.date),
            self.WEEKEND_CURVE.eval(self.date),
            self.GROWTH_CURVE.eval(self.date),
        ]

    def at_minute(self, minutes):
        return Day(self.date_index, minutes=minutes)

    def get_effect(self):
        total = 1
        for effect in self.effects:
            total = total * effect
        return total

        # weekend_effect = 0.8 if date.is_weekend else 1
        # summer_effect = 0.7 if date.season == 'summer' else 1

    @property
    def day_of_week(self):
        return self.date.weekday()

    @property
    def is_weekend(self):
        # 5 + 6 are weekends
        return self.date.weekday() >= 5

    @property
    def season(self):
        month_no = self.date.month
        day_no = self.date.day

        if month_no in (1, 2) or (month_no == 3 and day_no < 21):
            return "winter"
        elif month_no in (3, 4, 5) or (month_no == 6 and day_no < 21):
            return "spring"
        elif month_no in (6, 7, 8) or (month_no == 9 and day_no < 21):
            return "summer"
        elif month_no in (9, 10, 11) or (month_no == 12 and day_no < 21):
            return "fall"
        else:
            return "winter"



import random
import uuid


class Tweet:
    def __init__(self, customer, tweet_time, order) -> None:
        self.uuid = str(uuid.uuid4())
        self.tweeted_at = tweet_time
        self.customer = customer
        self.order = order
        self.content = self.construct_tweet()

    def to_dict(self) -> dict[str, str]:
        return {
            "id": self.uuid,
            "user_id": self.customer.customer_id,
            "tweeted_at": str(self.tweeted_at.isoformat()),
            "content": self.content,
        }

    def construct_tweet(self) -> str:
        if len(self.order.items) == 1:
            items_sentence = f"Ordered a {self.order.items[0].item.name}"
        elif len(self.order.items) == 2:
            items_sentence = f"Ordered a {self.order.items[0].item.name} and a {self.order.items[1].item.name}"
        else:
            items_sentence = f"Ordered a {', a '.join(item.item.name for item in self.order.items[:-1])}, and a {self.order.items[-1].item.name}"
        if self.customer.fan_level > 3:
            adjective = random.choice(
                [
                    "the best",
                    "awesome",
                    "delicious",
                    "amazing",
                    "fantastic",
                    "sooo gooood",
                    "my favorite",
                ]
            )
            return f"Jaffles from the Jaffle Shop are {adjective}! {items_sentence}."
        elif self.customer.fan_level < 3:
            adjective = random.choice(
                [
                    "terrible",
                    "the worst",
                    "awful",
                    "disgusting",
                    "gross",
                    "inedible",
                    "my least favorite",
                ]
            )
            return f"Jaffle Shop again. {items_sentence}. This place is {adjective}."
        else:
            adjective = random.choice(
                [
                    "okay",
                    "fine",
                    "alright",
                    "average",
                    "pretty decent",
                    "solid",
                    "not bad",
                    "just meh",
                ]
            )
            return f"Jaffle shop is {adjective}. {items_sentence}."



import random
import datetime
import uuid
from abc import ABC, abstractmethod
from typing import Any

import numpy as np
from faker import Faker

from jafgen.customers.order import Order
from jafgen.customers.tweet import Tweet
from jafgen.stores.inventory import Inventory

fake = Faker()
Faker.seed(123456789)


class Customer(ABC):
    def __init__(self, store):
        self.customer_id = str(uuid.uuid4())
        self.store = store
        self.name = fake.name()
        self.favorite_number = int(np.random.rand() * 100)
        self.fan_level = random.randint(1, 5)

    def p_buy_season(self, day):
        return self.store.p_buy(day)

    def p_buy(self, day):
        p_buy_season = self.p_buy_season(day)
        p_buy_persona = self.p_buy_persona(day)
        p_buy_on_day = (p_buy_season * p_buy_persona) ** 0.5
        return p_buy_on_day

    def p_tweet(self, day):
        return self.p_tweet_persona(day)

    def get_order(self, day):
        items = self.get_order_items(day)
        order_time_delta = self.get_order_time(day)

        order_time = day.at_minute(order_time_delta + self.store.opens_at(day))
        if not self.store.is_open_at(order_time):
            return None

        return Order(self, items, self.store, order_time)

    def get_tweet(self, order):
        tweet_time = order.order_time.date + datetime.timedelta(
            minutes=(np.random.rand() * 20)
        )
        return Tweet(self, tweet_time, order)

    @abstractmethod
    def get_order_items(self, day):
        raise NotImplementedError()

    @abstractmethod
    def get_order_time(self, day):
        raise NotImplementedError()

    @abstractmethod
    def p_buy_persona(self, day):
        raise NotImplementedError()

    @abstractmethod
    def p_tweet_persona(self, day):
        raise NotImplementedError()

    def sim_day(self, day):
        p_buy = self.p_buy(day)
        p_buy_threshold = np.random.random()
        p_tweet = self.p_tweet(day)
        p_tweet_threshold = np.random.random()
        if p_buy > p_buy_threshold:
            if p_tweet > p_tweet_threshold:
                order = self.get_order(day)
                if order and len(order.items) > 0:
                    return order, self.get_tweet(order)
                else:
                    return None, None
            else:
                return self.get_order(day), None
        else:
            return None, None

    def to_dict(self) -> dict[str, Any]:
        return {
            "id": str(self.customer_id),
            "name": str(self.name),
        }


class RemoteWorker(Customer):
    "This person works from a coffee shop"

    def p_buy_persona(self, day):
        buy_propensity = (self.favorite_number / 100) * 0.4
        return 0.001 if day.is_weekend else buy_propensity

    def p_tweet_persona(self, day):
        return 0.01

    def get_order_time(self, day):
        # most likely to order in the morning
        # exponentially less likely to order in the afternoon
        avg_time = 60 * 7
        order_time = np.random.normal(loc=avg_time, scale=180)
        return max(0, int(order_time))

    def get_order_items(self, day):
        num_drinks = 1
        food = []

        if random.random() > 0.7:
            num_drinks = 2

        if random.random() > 0.7:
            food = Inventory.get_food(1)

        return Inventory.get_drink(num_drinks) + food


class BrunchCrowd(Customer):
    "Do you sell mimosas?"

    def p_buy_persona(self, day):
        buy_propensity = 0.2 + (self.favorite_number / 100) * 0.2
        return buy_propensity if day.is_weekend else 0

    def p_tweet_persona(self, day):
        return 0.8

    def get_order_time(self, day):
        # most likely to order in the early afternoon
        avg_time = 300 + ((self.favorite_number - 50) / 50) * 120
        order_time = np.random.normal(loc=avg_time, scale=120)
        return max(0, int(order_time))

    def get_order_items(self, day):
        num_customers = 1 + int(self.favorite_number / 20)
        return Inventory.get_drink(num_customers) + Inventory.get_food(num_customers)


class Commuter(Customer):
    "the regular, thanks"

    def p_buy_persona(self, day):
        buy_propensity = 0.5 + (self.favorite_number / 100) * 0.3
        return 0.001 if day.is_weekend else buy_propensity

    def p_tweet_persona(self, day):
        return 0.2

    def get_order_time(self, day):
        # most likely to order in the morning
        # exponentially less likely to order in the afternoon
        avg_time = 60
        order_time = np.random.normal(loc=avg_time, scale=30)
        return max(0, int(order_time))

    def get_order_items(self, day):
        return Inventory.get_drink(1)


class Student(Customer):
    "coffee might help"

    def p_buy_persona(self, day):
        if day.season == "summer":
            return 0
        else:
            buy_propensity = 0.1 + (self.favorite_number / 100) * 0.4
            return buy_propensity

    def p_tweet_persona(self, day):
        return 0.8

    def get_order_time(self, day):
        # later is better
        avg_time = 9 * 60
        order_time = np.random.normal(loc=avg_time, scale=120)
        return max(0, int(order_time))

    def get_order_items(self, day):
        food = []
        if random.random() > 0.5:
            food = Inventory.get_food(1)

        return Inventory.get_drink(1) + food


class Casuals(Customer):
    "just popping in"

    def p_buy_persona(self, day):
        return 0.1

    def p_tweet_persona(self, day):
        return 0.1

    def get_order_time(self, day):
        avg_time = 5 * 60
        order_time = np.random.normal(loc=avg_time, scale=120)
        return max(0, int(order_time))

    def get_order_items(self, day):
        num_drinks = int(random.random() * 10 / 3)
        num_food = int(random.random() * 10 / 3)
        return Inventory.get_drink(num_drinks) + Inventory.get_food(num_food)


class HealthNut(Customer):
    "A light beverage in the sunshine as a treat"

    def p_buy_persona(self, day):
        if day.season == "summer":
            buy_propensity = 0.1 + (self.favorite_number / 100) * 0.4
            return buy_propensity
        else:
            return 0.2

    def p_tweet_persona(self, day):
        return 0.6

    def get_order_time(self, day):
        avg_time = 5 * 60
        order_time = np.random.normal(loc=avg_time, scale=120)
        return max(0, int(order_time))

    def get_order_items(self, day):
        return Inventory.get_drink(1)



import uuid
from typing import Any

from jafgen.customers.order_item import OrderItem


class Order(object):
    def __init__(self, customer, items, store, order_time):
        self.order_id = str(uuid.uuid4())
        self.customer = customer
        self.items = [OrderItem(self.order_id, item) for item in items]
        self.store = store
        self.order_time = order_time
        self.subtotal = sum(i.item.price for i in self.items)
        self.tax_paid = store.tax_rate * self.subtotal
        self.order_total = self.subtotal + self.tax_paid

    def __str__(self):
        return f"{self.customer.name} bought {str(self.items)} at {self.order_time}"

    def to_dict(self) -> dict[str, Any]:
        return {
            "id": str(self.order_id),
            "customer": str(self.customer.customer_id),
            "ordered_at": str(self.order_time.date.isoformat()),
            "store_id": str(self.store.store_id),
            "subtotal": int(self.subtotal * 100),
            "tax_paid": int(self.tax_paid * 100),
            # TODO: figure out why this is doesn't cause a test failure
            # in tests/test_order_totals.py
            # "order_total": int(self.order_total * 100),
            "order_total": int(int(self.subtotal * 100) + int(self.tax_paid * 100)),
        }

    def items_to_dict(self) -> list[dict[str, Any]]:
        return [item.to_dict() for item in self.items]



from typing import Annotated
import typer

from jafgen.simulation import Simulation

app = typer.Typer()


@app.command()
def run(
    years: Annotated[
        int, typer.Argument(help="Number of years to simulate. Default is 1.")
    ] = 1,
    pre: Annotated[
        str,
        typer.Option(help="Optional prefix for the output file names."),
    ] = "raw",
):
    sim = Simulation(years, pre)
    sim.run_simulation()
    sim.save_results()



from setuptools import find_packages, setup

setup(
    name="jafgen",
    version="0.4.13",
    packages=find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests"],
    ),
    authors=["gwenwindflower", "drewbanin"],
    description="A synthetic data generator CLI for a fictional Jaffle Shop",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    package_dir={"jafgen": "jafgen"},
    entry_points={"console_scripts": ["jafgen = jafgen.cli:app"]},
    install_requires=["numpy", "Faker", "typer"],
)



from jafgen.stores.inventory import Inventory
from jafgen.stores.stock import Stock
from jafgen.stores.supply import Supply


def test_stock_and_inventory_equal():
    """Ensure Stock and Inventory have an equal number of items"""
    assert len(Stock.stock) == len(Inventory.inventory["jaffle"]) + len(
        Inventory.inventory["beverage"]
    )


def test_all_inventory_beverages_are_beverages():
    """Ensure all the items in inventory.beverage are of type beverage"""
    for item in Inventory.inventory["beverage"]:
        assert item.type == "beverage"


def test_all_inventory_jaffles_are_jaffles():
    """Ensure all the items in inventory.jaffle are of type jaffle"""
    for item in Inventory.inventory["jaffle"]:
        assert item.type == "jaffle"


def test_inventory_stock_all_has_supplies():
    """Ensure all Supplies have the necessary properties"""
    for menu_item in Stock.stock:
        for supply in Stock.stock[menu_item]:
            assert isinstance(supply, Supply)


def test_supplies_have_all_props():
    """Ensure all Supplies have the necessary properties"""
    supply_attrs = ["id", "name", "cost", "perishable", "skus"]
    for menu_item in Stock.stock:
        for supply in Stock.stock[menu_item]:
            for attr in supply_attrs:
                assert hasattr(supply, attr)
